from allwhois import whois


class TestUnicodeDomains:

    def test_no_match(self):
        domain = "нарояци.com"
        result = whois.query(domain)
        assert result["msg"] == "No match"

    def test_parsing_error(self):
        domain = "öbb.at"
        result = whois.query(domain)
        assert result["msg"] == "Err parsing"




"""
cache raw outputs to file
"""
import collections as _collections
import datetime as _datetime
import functools as _functools
import inspect as _inspect
import os as _os
import pickle as _pickle
import codecs as _codecs
import shelve as _shelve
import sys as _sys
import time as _time
import traceback as _traceback
import types
import atexit

_retval = _collections.namedtuple('_retval', 'timesig data')
_SRC_DIR = _os.path.dirname(_os.path.abspath(__file__))
_CACHE_STORE = '.cache'

SECOND = 1
MINUTE = 60 * SECOND
HOUR = 60 * MINUTE
DAY = 24 * HOUR
WEEK = 7 * DAY
MONTH = 30 * DAY
YEAR = 365 * DAY
FOREVER = None

OPEN_DBS = dict()


def _get_cache_name(function):
    """
    returns a name for the module's cache db.
    """
    module_name = _inspect.getfile(function)
    cache_name = _os.path.basename(_os.path.splitext(module_name)[0])
    if not _os.path.exists(_CACHE_STORE):
        _os.makedirs(_CACHE_STORE)

    # fix for '<string>' or '<stdin>' in exec or interpreter usage.
    cache_name = cache_name.replace('<', '').replace('>', '')
    cache_name = _os.path.join(_CACHE_STORE, f'{cache_name}.cache')

    return cache_name


def _log_error(error_str):
    try:
        error_log_fname = _os.path.join(_SRC_DIR, 'filecache.err.log')
        if _os.path.isfile(error_log_fname):
            fhand = open(error_log_fname, 'a')
        else:
            fhand = open(error_log_fname, 'w')
        fhand.write('[%s] %s\r\n' % (_datetime.datetime.now().isoformat(), error_str))
        fhand.close()
    except Exception:
        pass


def _args_key(function, args, kwargs):
    arguments = (args, kwargs)
    # Check if you have a valid, cached answer, and return it.
    if _sys.version_info[0] == 2:
        arguments_pickle = _pickle.dumps(arguments)
    else:
        # NOTE: protocol=0 so it's ascii, this is crucial for py3k
        #       because shelve only works with proper strings.
        #       Otherwise, we'd get an exception because
        #       function.__name__ is str but dumps returns bytes.
        arguments_pickle = _codecs.encode(_pickle.dumps(arguments, protocol=0), "base64").decode()

    key = function.__name__ + arguments_pickle
    return key


def filecache(seconds_of_validity=None, fail_silently=False):
    """
    filecache is called and the decorator should be returned.
    """

    def filecache_decorator(function):
        @_functools.wraps(function)
        def function_with_cache(*args, **kwargs):
            key = None
            try:
                key = _args_key(function, args, kwargs)

                if key in function._db:
                    rv = function._db[key]
                    if seconds_of_validity is None or _time.time() - rv.timesig < seconds_of_validity:
                        return rv.data
            except Exception:
                # in any case of failure, don't let filecache break the program
                error_str = _traceback.format_exc()
                _log_error(error_str)
                if not fail_silently:
                    raise

            retval = function(*args, **kwargs)

            # store in cache
            # NOTE: no need to _db.sync() because there was no mutation
            # NOTE: it's importatnt to do _db.sync() because otherwise the cache doesn't survive Ctrl-Break!
            try:
                function._db[key] = _retval(_time.time(), retval)
                function._db.sync()
            except Exception:
                # in any case of failure, don't let filecache break the program
                error_str = _traceback.format_exc()
                _log_error(error_str)
                if not fail_silently:
                    raise

            return retval

        # make sure cache is loaded
        if not hasattr(function, '_db'):
            cache_name = _get_cache_name(function)
            if cache_name in OPEN_DBS:
                function._db = OPEN_DBS[cache_name]
            else:
                function._db = _shelve.open(cache_name)
                OPEN_DBS[cache_name] = function._db
                atexit.register(function._db.close)

            function_with_cache._db = function._db

        return function_with_cache

    if isinstance(seconds_of_validity, types.FunctionType):
        # support when used as '@filecache.filecache' instead of '@filecache.filecache()'
        func = seconds_of_validity
        seconds_of_validity = None
        return filecache_decorator(func)

    return filecache_decorator




"""
Normaliser
"""
__all__ = ["Normaliser", "_dedupe_list"]

import calendar
import functools
import json
import operator
import re
import string
import time
from datetime import datetime, timedelta
from typing import Union, List

template0 = {
    # Unified Key: [Keys from various sources]
    'domain_name': ['domainName', 'domain', 'domainname'],
    'domain_canonical_name': ['canonical_name'],
    'domain_created_date': ['domain_name_commencement_date', 'domain_created', 'createDate', 'createdDate', 'connected_date', 'registered', 'registered_date', 'created_on', 'registered_on', 'registration_time', 'domain_record_activated', 'entry_created', 'create_date', 'creation_date', 'record_created'],
    'domain_expiry_date': ['validity', 'valid_until', 'expiration_date', 'expire', 'expire_date', 'renewal_date', 'expiry_date', 'domain_expires', 'expired_date', 'paid_till', 'expires', 'expiration_time', 'expires_on', 'expiresDate', 'registry_expiry_date', 'expires_date', 'record_expires_on'],
    'domain_updated_date': ['updated_on', 'last_updated_on', 'updateDate', 'updatedDate', 'update_date', 'updated', 'entry_updated', 'domain_record_last_updated', 'modified', 'last_updated', 'updated_date', 'last_update', 'last_modified', 'record_last_updated_on'],
    'domain_name_servers': ['dns', 'servers', 'nserver', 'domain_servers', 'name_servers', 'name_server_', 'domain_servers_in_listed_order', 'name_servers_in_the_listed_order', 'name_server_information', 'nameserver', 'nameservers', 'name_server', 'dns_servers', 'name_servers_information'],
    'domain_status': ['epp_status', 'flags', 'domainStatus', 'registration_status', 'state', 'domain_status_', 'status', 'domain_status'],
    'contact_email': ['contactEmail'],
    'registrant_name': ['domain_owner', 'registrant', 'registrant_contact', 'organization', 'owner', 'owner_name'],
    'registrant_id': ['ownerid', 'org_id', 'owner_id'],
    'registrant_country_code': ['owner_country_code'],
    'registrant_email': ['owner_email'],
    'registrant_locality': ['owner_locality'],
    'registrant_locality_zipcode': ['owner_locality_zipcode'],
    'registrant_zipcode': ['owner_zipcode'],
    'registrant_type': ['organization_type'],
    'registrant_address': ['address_loc', 'descr', 'owner_address'],
    'registrar_name': ['domain_registrar_name', 'registrar', 'sponsoring_registrar_name', 'registrarName'],
    'registrar_id': ['sponsoring_registrar_pandi_id', 'sponsoring_registrar_id'],
    'registrar_city': ['sponsoring_registrar_city'],
    'registrar_state_province': ['sponsoring_registrar_state_province'],
    'registrar_organization': ['sponsoring_registrar_organization'],
    'registrar_zip': ['sponsoring_registrar_postal_code'],
    'registrar_country': ['sponsoring_registrar_country'],
    'registrar_telephone': ['sponsoring_registrar_phone'],
    'registrar_email': ['sponsoring_registrar_contact_email'],
    'registrar_whois_server': ['whoisServer', 'domain_registrar_whois'],
    'registrar_whois_url': ['domain_registrar_url'],
    'registrar_iana_id': ['IANAID', 'domain_registrar_id', 'registrarIANAID'],
    'admin_contact': ['administrative_contact', 'admin_c'],
    'tech_contact': ['technical_contact', 'tech_c', 'technical_contact_information'],
    'zone_contact': ['zone_c'],
    'domain_dnssec': ['dnssec'],
    'domain_dnssec_ds_data': ['dnssec_ds_data'],
}

sub_keys_template0 = {
    # sub keys
    'zip': ['postalCode', 'postal_code', 'zipcode'],
    'locality_zip': ['locality_zipcode'],
    'fax_ext': ['faxExt', 'fax_ext'],
    'telephone': ['phone'],
    'telephone_ext': ['telephoneExt', 'telephone_ext', 'phone_ext'],
    'address': ['street_address', 'address', 'contact'],
    'organization': ['company'],
}

dnssec_codes = [
    "no", "unsigned", "signedDelegation", "unsigned delegation"
]

epp_codes = {
    'active': ['ACTIVE'],
    'inactive': ['INACTIVE'],
    'ok': ['OK'],
    'linked': ['LINKED'],
    'addPeriod': ['ADD PERIOD'],
    'autoRenewPeriod': ['AUTO RENEW PERIOD', 'AUTORENEWPERIOD'],
    'pendingCreate': ['PENDING CREATE', 'PENDINGCREATE'],
    'pendingDelete': ['PENDING DELETE', 'PENDINGDELETE'],
    'pendingRenew': ['PENDING RENEW', 'PENDINGRENEW'],
    'pendingRestore': ['PENDING RESTORE', 'PENDINGRESTORE'],
    'pendingUpdate': ['PENDING UPDATE', 'PENDINGUPDATE'],
    'redemptionPeriod': ['REDEMPTION PERIOD', 'REDEMPTIONPERIOD'],
    'renewPeriod': ['RENEW PERIOD', 'RENEWPERIOD'],
    'updateProhibited': ['Update forbidden'],
    'serverDeleteProhibited': ['SERVER DELETE PROHIBITED', 'SERVERDELETEPROHIBITED', 'Deletion forbidden'],
    'serverHold': ['SERVER HOLD', 'SERVERHOLD'],
    'serverRenewProhibited': ['SERVER RENEW PROHIBITED'],
    'serverTransferProhibited': ['SERVER TRANSFER PROHIBITED'],
    'serverUpdateProhibited': ['SERVER UPDATE PROHIBITED', 'Sponsoring registrar change forbidden'],
    'transferPeriod': ['TRANSFER PERIOD'],
    'transferLocked': ['TRANSFER LOCKED'],
    'clientHold': ['CLIENT HOLD'],
    'clientDeleteProhibited': ['CLIENT DELETE PROHIBITED', 'DELETE PROHIBITED'],
    'clientTransferProhibited': ['CLIENT TRANSFER PROHIBITED', 'TRANSFER PROHIBITED'],
    'clientUpdateProhibited': ['CLIENT UPDATE PROHIBITED', 'UPDATE PROHIBITED', 'Registrant change forbidden'],
    'clientRenewProhibited': ['CLIENT RENEW PROHIBITED', 'RENEW PROHIBITED'],
    'Administratively blocked': ['Administratively blocked']
}
inverted_epp_dict = {val: key for key, arr in epp_codes.items() for val in arr}
epp_re = re.compile(
    "(" + '|'.join(list(inverted_epp_dict.keys())) + "|" + '|'.join(list(inverted_epp_dict.values())) + ")",
    re.MULTILINE | re.IGNORECASE
)

pop_these = [
    "query", "reg_name", "data_validation", "relevant_dates", "regnr", "disclaimer", "holder", "nsset",
    "connected_date", "changed", "free_date", "person", "fax_no", "phone", "nic_hdl", "address", "url",
    "source", "mnt_by", "dom_public", "country_loc", "organization_loc", "person_loc", "postal_code_loc",
    "created", "nic_hdl_br", "nslastaa", "nsstat", "owner_c", "outzone", "delete", "more_information_at_http",
    "pid", "in_zone", "register_your_domain_name_at_https", "notice", "under_the_terms_and_conditions_at_https",
    "idn_tag", "family_name", "given_name", "reseller", "status_information", "re_registration_status"
]

DATE_FORMATS = [
    # http://docs.python.org/library/datetime.html#strftime-strptime-behavior
    '%Y-%b-%d',                     # 2000-jan-02
    '%d-%B-%Y',                     # 02-january-2000
    '%d-%b-%Y',                     # 02-jan-2000
    '%d-%b-%y',                     # 02-jan-00
    '%d.%m.%Y',                     # 02.02.2000
    '%d/%m/%Y',                     # 01/06/2011
    '%Y-%m-%d',                     # 2000-01-02
    '%d-%m-%Y',                     # 02-01-2000
    '%Y.%m.%d',                     # 2000.01.02
    '%Y/%m/%d',                     # 2005/05/30
    'before %b-%Y',                 # before aug-1996
    '%Y.%m.%d %H:%M:%S',            # 2002.09.19 13:00:00
    '%Y%m%d %H:%M:%S',              # 20110908 14:44:51
    '%Y-%m-%d %H:%M:%S',            # 2011-09-08 14:44:51
    '%Y-%m-%d %H:%M:%S CLST',       # 2011-09-08 14:44:51 CLST CL
    '%Y-%m-%d %H:%M:%S.%f',         # 2011-09-08 14:44:51 CLST CL
    '%d.%m.%Y %H:%M:%S',            # 19.09.2002 13:00:00
    '%d-%b-%Y %H:%M:%S %Z',         # 24-Jul-2009 13:20:03 UTC
    '%d/%m/%Y %H:%M:%S',            # 30/10/2002 00:00:00
    '%Y/%m/%d %H:%M:%S (%z)',       # 2011/06/01 01:05:01 (+0900)
    '%Y-%m-%d %H:%M:%S (%z)',       # 2011/06/01 01:05:01 (+0900)
    '%Y/%m/%d %H:%M:%S%z',          # 2011/06/01 01:05:01+0900
    '%Y-%m-%d %H:%M:%S%z',          # 2011-06-01 01:05:01+0900
    '%Y/%m/%d %H:%M:%S %z',         # 2011/06/01 01:05:01 +0900
    '%Y-%m-%d %H:%M:%S %z',         # 2011-06-01 01:05:01 +0900
    '%Y/%m/%d %H:%M:%S:%f%z',       # 2011/06/01 01:05:01:00+0900
    '%Y-%m-%d %H:%M:%S:%f%z',       # 2011-06-01 01:05:01:00+0900
    '%Y/%m/%d %H:%M:%S.%f%z',       # 2011/06/01 01:05:01.00+0900
    '%Y-%m-%d %H:%M:%S.%f%z',       # 2011-06-01 01:05:01.00+0900
    '%Y/%m/%d %H:%M:%S',            # 2011/06/01 01:05:01
    '%a %b %d %H:%M:%S %Z %Y',      # Tue Jun 21 23:59:59 GMT 2011
    '%a %b %d %H:%M:%S %Y',         # tue feb 10 09:42:42 2004
    '%a %b %d %Y',                  # Tue Dec 12 2000
    '%Y-%m-%dT%H:%M:%S',            # 2007-01-26T19:10:31
    '%Y-%m-%dT%H:%M:%SZ',           # 2007-01-26T19:10:31Z
    '%Y-%m-%dt%H:%M:%S.%fz',        # 2007-01-26t19:10:31.00z
    '%Y-%m-%dT%H:%M:%S%z',          # 2011-03-30T19:36:27+0200
    '%Y-%m-%dT%H:%M:%S.%f%z',       # 2011-09-08T14:44:51.622265+03:00
    '%Y-%m-%dt%H:%M:%S.%f',         # 2011-09-08t14:44:51.622265
    '%Y-%m-%dt%H:%M:%S',            # 2007-01-26T19:10:31
    '%Y-%m-%dt%H:%M:%SZ',           # 2007-01-26T19:10:31Z
    '%Y-%m-%dt%H:%M:%S.%fz',        # 2007-01-26t19:10:31.00z
    '%Y-%m-%dt%H:%M:%S%z',          # 2011-03-30T19:36:27+0200
    '%Y-%m-%dt%H:%M:%S.%f%z',       # 2011-09-08T14:44:51.622265+03:00
    '%Y%m%d',                       # 20110908
    '%Y. %m. %d.',                  # 2020. 01. 12.
    'before %b-%Y',                 # before aug-1996
    '%a %d %b %Y',                  # Tue 21 Jun 2011
    '%A %d %b %Y',                  # Tuesday 21 Jun 2011
    '%a %d %B %Y',                  # Tue 21 June 2011
    '%A %d %B %Y',                  # Tuesday 21 June 2011
    '%Y-%m-%d %H:%M:%S (%Z+0:00)',  # 2007-12-24 10:24:32 (gmt+0:00)
    '%B %d %Y',                     # January 01 2000
]
date_fields = [
    'domain_created_date', 'domain_updated_date', 'domain_expiry_date', 'registrar_registration_expiration_date'
]
contact_fields = [
    'registrant_', 'admin_', 'tech_', 'billing_', 'zone_'
]


def flatten(lst: list) -> List:
    """flatten the list of lists.
    :param lst: list of lists
    :return: flattened list
    """
    return functools.reduce(operator.iconcat, lst, [])


def _dedupe_list(lst: list) -> List:
    """case insensitive/fuzzy removal of duplicates or similarities from a list.
    :param lst: list of items
    :return: de-duplicated list of items
    """
    marker = set()
    lst = flatten(lst) if any(isinstance(el, list) for el in lst) else lst
    lst = list(filter(None, lst))
    result = [
        not marker.add(x.replace("www.", "").translate(
            str.maketrans('', '', string.punctuation)
        ).casefold()) and x
        for x in lst
        if x.replace("www.", "").translate(
            str.maketrans('', '', string.punctuation)
        ).casefold() not in marker
    ]
    return list(filter(None, result))


def _dedupe_dates(lst: list, precision: str = "day", tz_aware: bool = False) -> List:
    """Dedupe a given list of date objects
    :param lst: list of date objects
    :param precision: "day" to dedupe upto day (y/m/d) or "time" for y/m/d h:m:s
    :param tz_aware: If true, then dedupe keeping timezones information
    """
    dates_dict = {}
    k = None
    for dt in lst:
        if tz_aware:
            if not dt.tzinfo:
                timestamp = calendar.timegm(dt.timetuple())
                k = datetime.utcfromtimestamp(timestamp)
            else:
                timestamp = time.mktime(dt.timetuple())
                k = datetime.fromtimestamp(timestamp)
        else:
            if precision == "day":
                k = datetime(dt.year, dt.month, dt.day).timestamp()
            elif precision == "time":
                k = datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute, dt.second).timestamp()

        dates_dict.update({k: dt})

    return list(dates_dict.values())


class UnknownDateFormat(Exception):
    """Custom Exception."""
    pass


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, datetime.date, datetime.time)):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            return (datetime.min + obj).time().isoformat()

        return super(DateTimeEncoder, self).default(obj)


class Normaliser:

    date_as_string: bool = None

    def __setattr__(self, name, value):
        if name == 'date_as_string':
            super(Normaliser, self).__setattr__(name, value)

    @staticmethod
    def _group_keys(data: dict) -> dict:
        """Group certain keys together.
        :param data: dict to which the template needs to be applied
        :return: grouped keys dict
        """
        keys_to_group = (
            "registry_", "registrar_", "registrant_", "domain_", "admin_",
            "tech_", "billing_", "zone_", "option_", "company_"
        )
        for key in pop_these:
            # remove keys that are not necessary
            # they are still available in the z_raw key
            data.pop(key, None)

        # group the keys
        grouped = {}
        for key, value in data.items():
            if key.startswith(keys_to_group):
                new_key, sub_key = key.split("_", 1)
                if f"{new_key}_info" in grouped:
                    grouped[f"{new_key}_info"].update({sub_key: value})
                else:
                    grouped.update({f"{new_key}_info": {sub_key: value}})
            else:
                grouped.update({key: value})
        return grouped

    def _organise(self, data: dict) -> dict:
        """Organise & map the data appropriately.
        :param data: data dict
        :return: organised data dict
        """
        def _no_http(var):
            if var.startswith('http'):
                return False
            else:
                return True

        if 'domain_name' in data:
            if isinstance(data.get('domain_name'), list):
                data['domain_name'] = ', '.join(_dedupe_list(data.get('domain_name')))
        else:
            data['domain_name'] = ''
        if 'company_name' in data:
            if isinstance(data.get('company_name'), list):
                data['company_name'] = ', '.join(_dedupe_list(data.get('company_name')))
        if 'domain_dnssec' in data:
            if isinstance(data.get('domain_dnssec'), list):
                data['domain_dnssec'] = ', '.join(_dedupe_list(data.get('domain_dnssec')))
        if 'domain_name_servers' in data:
            if isinstance(data.get('domain_name_servers'), list):
                data['domain_name_servers'] = _dedupe_list(data.get('domain_name_servers'))
            elif isinstance(data.get('domain_name_servers'), str):
                data['domain_name_servers'] = data.get('domain_name_servers').split(' ')
            data['domain_name_servers'] = list(filter(None, data['domain_name_servers']))
            data['domain_name_servers'] = [
                ' ('.join([i.strip() for i in ns.split('\t')])+')'
                if '\t' in ns.strip() else ns
                for ns in data['domain_name_servers']
            ]
        if 'email' in data:
            if isinstance(data.get('email'), list):
                data['email'] = ', '.join(_dedupe_list(data.get('email')))
        if 'domain_status' in data:
            if isinstance(data.get('domain_status'), list):
                data['domain_status'] = flatten([i.split() for i in _dedupe_list(data.get('domain_status'))])
            elif isinstance(data.get('domain_status'), str):
                data['domain_status'] = data['domain_status'].split()
            data['domain_status'] = ' '.join(list(filter(_no_http, data['domain_status'])))
            data['domain_status'] = re.findall(epp_re, data['domain_status'])
        if 'domain_created_date' in data:
            if isinstance(data.get('domain_created_date'), list):
                data['domain_created_date'] = _dedupe_dates(data.get('domain_created_date'))
                if len(data['domain_created_date']) == 1:
                    data['domain_created_date'] = data['domain_created_date'][0]
        if 'domain_expiry_date' in data:
            if isinstance(data.get('domain_expiry_date'), list):
                data['domain_expiry_date'] = _dedupe_dates(data.get('domain_expiry_date'))
                if len(data['domain_expiry_date']) == 1:
                    data['domain_expiry_date'] = data['domain_expiry_date'][0]
        if 'domain_updated_date' in data:
            if isinstance(data.get('domain_updated_date'), list):
                data['domain_updated_date'] = _dedupe_dates(data.get('domain_updated_date'))
                if len(data['domain_updated_date']) == 1:
                    data['domain_updated_date'] = data['domain_updated_date'][0]
                elif len(data['domain_updated_date']) > 1:
                    data['domain_updated_date'] = max([str(d) for d in data['domain_updated_date']])
        if 'registrar_abuse_contact_email' in data:
            if isinstance(data.get('registrar_abuse_contact_email'), list):
                data['registrar_abuse_contact_email'] = ', '.join(_dedupe_list(data['registrar_abuse_contact_email']))
        if 'registrar_abuse_contact_phone' in data:
            if isinstance(data.get('registrar_abuse_contact_phone'), list):
                data['registrar_abuse_contact_phone'] = ', '.join(_dedupe_list(data['registrar_abuse_contact_phone']))
        if 'registrar_iana_id' in data:
            if isinstance(data.get('registrar_iana_id'), list):
                data['registrar_iana_id'] = ', '.join(_dedupe_list(data['registrar_iana_id']))
        if 'registrar_name' in data:
            if isinstance(data.get('registrar_name'), list):
                data['registrar_name'] = ', '.join(_dedupe_list(data['registrar_name']))
        if 'registrar_url' in data:
            if isinstance(data.get('registrar_url'), list):
                data['registrar_url'] = ', '.join(_dedupe_list(data['registrar_url']))
        if 'registrar_whois_server' in data:
            if isinstance(data.get('registrar_whois_server'), list):
                data['registrar_whois_server'] = ', '.join(_dedupe_list(data['registrar_whois_server']))
        if 'registry_domain_id' in data:
            if isinstance(data.get('registry_domain_id'), list):
                data['registry_domain_id'] = ', '.join(_dedupe_list(data['registry_domain_id']))
        if 'admin_street' in data:
            if isinstance(data.get('admin_street'), list):
                data['admin_street'] = ''.join(flatten(data.get('admin_street')))
        if 'billing_street' in data:
            if isinstance(data.get('billing_street'), list):
                data['billing_street'] = ''.join(flatten(data.get('billing_street')))
        if 'tech_street' in data:
            if isinstance(data.get('tech_street'), list):
                data['tech_street'] = ''.join(flatten(data.get('tech_street')))
        if 'registrant_street' in data:
            if isinstance(data.get('registrant_street'), list):
                data['registrant_street'] = ''.join(flatten(data.get('registrant_street')))
        if 'city' in data:
            if isinstance(data.get('city'), list):
                data['city'] = ', '.join(_dedupe_list(data['city']))
        if 'country' in data:
            if isinstance(data.get('country'), list):
                data['country'] = ', '.join(_dedupe_list(data['country']))

        if 'domain_created_date' not in data:
            data['domain_created_date'] = ''
        if 'domain_expiry_date' not in data:
            data['domain_expiry_date'] = ''
        if 'domain_updated_date' not in data:
            data['domain_updated_date'] = ''
        if 'domain_status' not in data:
            data['domain_status'] = []

        for addr in contact_fields:
            address = []
            postal = None
            if f'{addr}street' in data:
                if isinstance(data[f'{addr}street'], list):
                    data[f'{addr}street'] = ', '.join(_dedupe_list(data[f'{addr}street']))
                address.append(data[f'{addr}street'].strip())
            if f'{addr}state_province' in data:
                if isinstance(data[f'{addr}state_province'], list):
                    data[f'{addr}state_province'] = ', '.join(_dedupe_list(data[f'{addr}state_province']))
                address.append(data[f'{addr}state_province'].strip())
            if f'{addr}postal_code' in data:
                if isinstance(data[f'{addr}postal_code'], list):
                    data[f'{addr}postal_code'] = ', '.join(_dedupe_list(data[f'{addr}postal_code']))
                postal = data[f'{addr}postal_code'].strip()
            if f'{addr}country' in data:
                if isinstance(data[f'{addr}country'], list):
                    data[f'{addr}country'] = ', '.join(_dedupe_list(data[f'{addr}country']))
                address.append(f"{data[f'{addr}country'].strip()} {postal if postal else ''}".strip())
            if f'{addr}phone' in data:
                if isinstance(data[f'{addr}phone'], list):
                    data[f'{addr}phone'] = ', '.join(_dedupe_list(data[f'{addr}phone']))
                ph = data[f'{addr}phone'].strip()
                if ph:
                    address.append(f"Phone: {ph}")
            if f'{addr}fax' in data:
                if isinstance(data[f'{addr}fax'], list):
                    data[f'{addr}fax'] = ', '.join(_dedupe_list(data[f'{addr}fax']))
                fax = data[f'{addr}fax'].strip()
                if fax:
                    address.append(f"Fax: {fax}")

            if addr in '\t'.join(list(data.keys())):
                if f'{addr}address' not in data:
                    # if the address field is not there
                    data[f'{addr}address'] = ', '.join(list(filter(None, address)))
                elif not data[f'{addr}address']:
                    # if the address field is there, but empty
                    data[f'{addr}address'] = ', '.join(list(filter(None, address)))

        return data

    def _apply_template(self, data: dict, template: dict) -> dict:
        """Use a template to normalise/unify the keys of the data dict.
        :param data: input dict to which the template needs to be applied
        :param template: template dict to apply normalisation on data
        :return: template applied unified dict
        """
        result = {}
        for key, value in data.items():
            if isinstance(value, dict):
                result[key] = self._apply_template(value, template)
            else:
                new_key = ([x for x in template if key in template[x]] or [key])[0]
                result[new_key] = value

        return result

    def _standardise_dates(self, data: dict) -> dict:
        """Normalise all available date fields.
        :param data: data dict
        :return: date normalised data dict
        """
        def _tz_check(dt: str) -> str:
            match = [m.start() for m in re.finditer(r'\+|\-+', dt) if m.start() >= 15]
            if len(match) == 1:
                tz_len = len(dt[match[0]+1:])
                if tz_len <= 2:
                    dt = f"{dt}00"
            return dt

        def _to_std_date(val: str) -> Union[str, datetime]:
            val = val.lower().replace('(jst)', '(+0900)')
            val = re.sub(r"(\+[0-9]{2}):([0-9]{2})", r"\1\2", val)
            val = re.sub(r"(\ #.*)", "", val)
            val = re.sub(r"(\d+)(st|nd|rd|th) ", r"\1 ", val)
            val = _tz_check(val)
            for fmt in DATE_FORMATS:
                try:
                    _standardised_date = datetime.strptime(val, fmt)
                    return _standardised_date
                except ValueError:
                    pass

            raise UnknownDateFormat(f"Unknown date format: '{val}'")

        for field in date_fields:
            if field in data:
                if isinstance(data[field], list):
                    data[field] = [_to_std_date(dt) for dt in data[field]]
                elif isinstance(data[field], str):
                    if data[field]:
                        data[field] = _to_std_date(data[field])

        return data

    def normalise(self, dossier: dict):
        """Normalisation of the given input data.
        :param dossier: input data
        :return:
        """
        # normalise the data dict with unified keys
        normalised = self._apply_template(dossier, template0)
        # normalise all the date fields
        date_normalised = self._standardise_dates(normalised)
        # organise the data
        organised = self._organise(date_normalised)
        # group the keys in the dict
        normalised = self._group_keys(organised)
        # reapply template for the grouped sub-keys
        normalised = self._apply_template(normalised, sub_keys_template0)

        if self.date_as_string:
            normalised = json.loads(DateTimeEncoder().encode(normalised))

        return normalised



from allwhois import whois
from .results import lookup_results


class TestDomainLookups:

    def test_nonexistant_tld(self):
        domain = "nonexistant.tld"
        result = whois.query(domain, date_as_string=True)
        assert result == lookup_results['nonexistant_tld']

    def test_google_com(self):
        domain = "google.com"
        result = whois.query(domain, date_as_string=True)
        assert result["msg"] == "success"
        assert result["domain_info"]["name"].lower() == "google.com"
        assert result["domain_info"] == lookup_results['google_com']["domain_info"]

    def test_google_net(self):
        domain = "google.net"
        result = whois.query(domain, date_as_string=True)
        assert result["msg"] == "success"
        assert result["domain_info"]["name"].lower() == "google.net"
        assert result["domain_info"] == lookup_results['google_net']["domain_info"]

    def test_google_org(self):
        domain = "google.org"
        result = whois.query(domain, date_as_string=True)
        assert result["msg"] == "success"
        assert result["domain_info"]["name"].lower() == "google.org"
        assert result["domain_info"] == lookup_results['google_org']["domain_info"]

    def test_example_com(self):
        domain = "example.com"
        result = whois.query(domain, date_as_string=True)
        assert result["msg"] == "success"
        assert result["domain_info"]["name"].lower() == "example.com"
        assert result["domain_info"] == lookup_results['example_com']["domain_info"]




from allwhois import whois
from datetime import datetime


class TestDates:

    def test_datetime_object(self):
        domain = "google.com"
        result = whois.query(domain)
        assert isinstance(result["domain_info"]["created_date"], datetime)

    def test_datetime_as_string(self):
        domain = "google.com"
        result = whois.query(domain, date_as_string=True)
        assert isinstance(result["domain_info"]["created_date"], str)



"""
Parse the text output
"""
__all__ = ["Parser"]

import re
from textwrap import dedent
from typing import Union

from .normalise import _dedupe_list

_patterns = {
    'delim_generic': re.compile(
        r"^(Last updated on|"
        r"Technical Contacts|"
        r"Admin Contact|"
        r"Nameservers|"
        r"Registrant|"
        r"Registrar|"
        r"address|"
        r"DNSSEC|"
        r"Holder|"
        r"Tech)"
        r"[^:]*$",
        re.MULTILINE | re.IGNORECASE
    ),
    'delim_sq': re.compile(r"(^(?:[a-z]\.\s+|)\[[^]]+\])", re.MULTILINE),
    'alpha_start': re.compile(r"^(\w\.\s)\["),
    'star_start': re.compile(f"(^\*+\s*[\w\-\/\t\.\(\)\\ ]+)(?=\:):", re.MULTILINE),
    'dotted_end': re.compile(r"([\w\-]+\.[\w\-]+\.[\w\-]+\.[\s$]|\:.*\.$)", re.MULTILINE),
    'non_email_val': re.compile(r"^(.*: please query the .*)\n?", re.MULTILINE | re.IGNORECASE),
    'ipv4or6': re.compile(r"([0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3})|(([a-f0-9:]+:+)+[a-f0-9]+)"),
    'field': re.compile(r"(^[\w\-\/\t\'\.\(\) ]+)(?=\:):(?![0-9a-fA-F]{1,4}\:)", re.MULTILINE | re.IGNORECASE),
    'alt_field': re.compile(r"(^[^\t][\w\-\/\t\'\.\(\) ]+)[\\t]*(?=\:):(?![0-9a-fA-F]{1,4}\:)", re.MULTILINE | re.IGNORECASE),
    'key': re.compile(r"\s\w*~")
}
discard_prefix = (
    '#', "*", "-", "%", ">", "==", "url of", "for more information", "http", "web-based",
    "visit", "contact us", "protecting", "information", "whois ", "copyright", "in ",
    "markmonitor domain management", "protection according", "rules of", "estonia .ee"
)
discard_suffix = (
    ".", ","
)
discard_words = (
    "agreement",
)


class Parser:

    @staticmethod
    def _scrub(raw: str) -> str:
        """Scrub & Clean the given Raw Text.

        (1) remove paragraphs
        (2) remove indents
        (3) check for servers/server names ending with . and trim it
        (4) check for non email value in email field and remove it
        (5) check for fields not ending with delim ":" and add it
        (6) check for fields in square brackets and convert to normal fields
        :param raw: raw text as input
        :return: cleaned output as text string
        """
        def _word_count(string: str) -> int:
            """count the number of words in the given string."""
            return len(string.split())

        def _rm_dot(match: re.Match) -> str:
            """if any potential field having a . at the end."""
            if match.group(1).strip()[-1] == '.':
                val = match.group(1).strip()[:-1]
            else:
                val = match.group(1)
            return f'{val}'

        def _rm_star(match: re.Match) -> str:
            """if any potential field having a * at the beginning."""
            val = match.group(1)
            if val.startswith('*'):
                val = val.replace('*', '').strip()
            else:
                val = val[:-1]
            return f"{val}:"

        def _rm_delim_sq(match: re.Match) -> str:
            """if any potential field is surrounded with []."""
            val = match.group(1)
            if re.findall(_patterns.get('alpha_start'), val):
                # trim fields start with char. (eg: a. )
                val = val[2:].strip()
            val = val.replace('[', '').replace(']', '')
            val = f"{val}:"
            return val

        def _non_eml_val(match: re.Match) -> str:
            """if we dont have an email in a email field."""
            val = match.group(1).split(':', 1)
            if val:
                val = f"{val[0]}: \n"
            else:
                val = match.group(1)
            return val

        # remove indent from lines
        raw = "\n".join([dedent(line) for line in raw.splitlines() if line])
        # remove a "." if its a valid field, so they dont get scrubbed away
        raw = re.sub(_patterns.get('dotted_end'), _rm_dot, raw)
        # remove "*" at the beginning if its a valid field, so they dont get scrubbed away
        raw = re.sub(_patterns.get('star_start'), _rm_star, raw)
        # remove invalid email addresses from email field
        raw = re.sub(_patterns.get('non_email_val'), _non_eml_val, raw)
        output = "\n".join(
            [
                # remove all of the unwanted lines that is not needed
                # and keep only the domain result "field: value"
                line.strip()
                for line in raw.splitlines()
                if not line.lower().strip().startswith(discard_prefix)
                and not line.lower().strip().endswith(discard_suffix)
                and not set(discard_words).intersection(line.lower().strip().split())
                and not _word_count(line) >= 10
            ]
        )
        # check if generic fields dont have ":" and add the ":"
        output = re.sub(_patterns.get('delim_generic'), r"\1:", output)
        # remove square brackets and add the ":" delim
        output = re.sub(_patterns.get('delim_sq'), _rm_delim_sq, output)

        return output

    @staticmethod
    def _fixes(data: dict) -> dict:
        if 'created' in data:
            # move created to createDate, as created has
            # conflicts with different domains outputs
            if isinstance(data['created'], list):
                # if its a list, the 0th element has the original created date
                data['createDate'] = data['created'][0]
            else:
                data['createDate'] = data['created']

        if 'changed' in data:
            if isinstance(data['changed'], list):
                # some of the whois output has assigned inside of changed field
                # which holds the creation date of the domain. Look for it and extract
                c_dt = []
                u_dt = []
                for dt in data['changed']:
                    if dt.lower().endswith('(assigned)'):
                        c_dt.append(re.findall(r"\s\d+\s", dt)[0].strip())
                        if c_dt:
                            break
                    if 'not disclosed' not in dt.lower() and 'whois' not in dt.lower()\
                            and '@' not in dt.lower():
                        # otherwise get the update date from "changed"
                        u_dt.append(dt)

                if 'created' not in data:
                    # created date found and created was not in the data
                    data['createDate'] = c_dt
                else:
                    # created date found, but update only if
                    # original created date in data is empty
                    if not data['created']:
                        data['createDate'] = c_dt

                # updated date
                data['updateDate'] = u_dt
            else:
                if 'created' not in data:
                    # created date found and created was not in the data
                    data['createDate'] = data['changed']
                else:
                    if not data['created']:
                        # created date found, but update only if
                        # original created date in data is empty
                        data['createDate'] = data['changed']
        if 'registered' in data:
            data.pop('createDate', None)
            data.pop('created', None)

        if 'descr' in data:
            if isinstance(data['descr'], list):
                data['descr'] = ' '.join(data['descr'])
                if 'phone' in data:
                    if isinstance(data['phone'], list):
                        if len(data['phone']) >= 3:
                            data['descr'] = f"{data['descr']}. {data['phone'][0]} (Phone)"
                            data['phone'] = data['phone'][1:]
                if 'fax_no' in data:
                    if isinstance(data['fax_no'], list):
                        if len(data['fax_no']) >= 3:
                            data['descr'] = f"{data['descr']}, {data['fax_no'][0]} (Fax)"
                            data['fax_no'] = data['fax_no'][1:]

        if 'address_loc' in data:
            if isinstance(data['address_loc'], list):
                data['address_loc'] = ' '.join(_dedupe_list(data['address_loc']))

        if 'e_mail' in data:
            if isinstance(data['e_mail'], list):
                data['email'] = ', '.join([elem.replace(' AT ', '@') for elem in _dedupe_list(data['e_mail'])])
                data.pop('e_mail', None)

        if 'fax' in data:
            if isinstance(data['fax'], list):
                data['fax'] = ', '.join(_dedupe_list(data['fax']))

        if 'name_server' in data and 'domain_servers_in_listed_order' in data:
            if isinstance(data['name_server'], list):
                data['name_server'].append(data['domain_servers_in_listed_order'])
                data.pop('domain_servers_in_listed_order', None)
                data['name_server'] = [e for e in data['name_server'] if e != '.']
            elif isinstance(data['domain_servers_in_listed_order'], list):
                data['domain_servers_in_listed_order'].append(data['name_server'])
                data.pop('name_server', None)
                data['domain_servers_in_listed_order'] = [e for e in data['domain_servers_in_listed_order'] if e != '.']
            else:
                data['name_server'] += data['domain_servers_in_listed_order']
                data.pop('domain_servers_in_listed_order', None)

        return data

    def _to_dict(self, text: str) -> dict:
        """Convert the cleaned text into dict.
        :param text: the cleaned raw text
        :return: dict
        """
        def _repl(match: re.Match) -> str:
            """replace extracted matched groups and convert to valid dict keys format."""
            val = match.group(1).strip().lower()
            if re.findall(r"^\d", val):
                # if the assumed field starts
                # with a number/digit; ignore it
                return match.group(1)+":"

            val = val.replace(
                " ", "_").replace(
                "-", "_").replace(
                "/", "_").replace(
                ".", "").replace(
                "[", "").replace(
                "]", "").replace(
                "'s", "")
            return f'{val}~'

        dict_o = {}
        inline = " " + " ".join(
            [
                re.sub(_patterns.get('alt_field'), _repl, line)
                for line in text.splitlines()
            ]
        ).replace('not.defined', '')

        # get all the keys
        keys = re.findall(_patterns.get('key'), inline)
        # create the kv pair
        kv = zip(
            [
                key.replace(' ', '').replace('~', '').strip()
                for key in keys
            ],
            re.sub('|'.join(keys), '|', inline).split('|')[1:]
        )
        # add the values of similar keys into a list value for the key
        for k, v in kv:
            if k in dict_o:
                if isinstance(dict_o[k], list) and v not in dict_o[k]:
                    dict_o[k].append(v.strip())
                else:
                    if dict_o[k] != v:
                        dict_o[k] = [dict_o[k]] + [v.strip()]
            else:
                dict_o[k] = v.strip()

        dict_o = self._fixes(dict_o)
        return dict_o

    def parse(self, text: str) -> Union[dict, None]:
        """Parse the raw whois output into a desirable pretty dict."""
        raw_text = self._scrub(text)
        if raw_text:
            pretty = self._to_dict(raw_text)
        else:
            pretty = None
        return pretty



"""
python wrapper for linux whois
Requirements: apt install whois (whois package)
"""
__all__ = ["Query"]

import subprocess
from shutil import which
from types import SimpleNamespace
from distutils.version import LooseVersion
from .parser import Parser
from .normalise import Normaliser
from ._filecache import filecache, DAY


class _NestedNamespace(SimpleNamespace):
    def __init__(self, dictionary, **kwargs):
        super().__init__(**kwargs)
        for key, value in dictionary.items():
            if isinstance(value, dict):
                self.__setattr__(key, _NestedNamespace(value))
            else:
                self.__setattr__(key, value)


class Query:
    executable: str = which('whois')
    encoding: str = 'utf-8'

    def __setattr__(self, name, value):
        if name in ('executable', 'encoding'):
            super(Query, self).__setattr__(name, value)

    @filecache(DAY*2)
    def _execute(self, args):
        """
        Calls out to subprocess with the passed in args
        This method is normally mocked in tests
        """
        try:
            output = subprocess.check_output(args, timeout=5)  # pragma: no cover
        except subprocess.CalledProcessError as grepexc:
            output = grepexc.output
        except subprocess.TimeoutExpired as t:
            output = "Timed out."
        try:
            output = output.decode(self.encoding)
        except AttributeError:
            output = None
        except:
            try:
                output = output.decode('ISO-8859-1')
            except:
                output = output.decode('latin')
        return output

    def _args(self, domain):
        """
        Builds up the final arguments to pass into subprocess
        whois <domain_name>
        """
        yield self.executable

        # suppress the licence from the output
        yield "-H"

        # The record we want to query
        yield domain

    @staticmethod
    def _preflight(domain):
        """Clean the input."""
        domain = domain.lower()
        domain = domain.replace('http://', '').replace('https://', '').strip()
        domain = domain.split('/', 1)[0]
        domain = domain.replace('www.', '')
        return domain

    def _validate_executable(self):
        """check if desired whois package is available"""
        if not self.executable:
            raise FileNotFoundError('whois is not found. install the whois package')
        ver = self._execute([self.executable, '--version'])
        if ver:
            ver = ver.split('\n')[0]
            if 'illegal option' in ver:
                raise FileNotFoundError('macos native whois not supported. install whois via brew')
            elif 'Version' in ver:
                ver_no = ver.split(' ', 1)[-1]
                if not LooseVersion(ver_no) > LooseVersion("5.5"):
                    raise ValueError("whois version should be >5.5, upgrade whois package")
        else:
            raise ValueError("whois version could not be determined, upgrade whois package")

    def to_dict(self):
        return self.__dict__

    def query(self, domain_name: str, date_as_string: bool = False):
        """
        :param domain_name: the domain name to collect whois information
        :param date_as_string: dates as string if true and not datetime format
        :return: a dict of whois info for the given domain
        """
        self._validate_executable()
        normaliser = Normaliser()

        domain = self._preflight(domain_name)
        args = list(self._args(domain))
        output = self._execute(args)
        if not output:
            result = {
                'domain_name': domain_name,
                'msg': 'fail'
            }
        elif "no match for " in output.lower():
            result = {
                'domain_name': domain_name,
                'msg': 'No match'
            }
        elif "timed out" in output.lower():
            result = {
                'domain_name': domain_name,
                'msg': 'Timeout error'
            }
        else:
            dict_o = Parser()
            result = dict_o.parse(output)
            if not result:
                # could not parse the output for some reason
                # scope to improve the parser
                result = {
                    'domain_name': domain_name,
                    'msg': 'Err parsing',
                    'z_raw': output
                }
            else:
                # final output
                result.update({
                    'msg': 'success',
                    'z_raw': output,
                })

                # normalise the output
                normaliser.date_as_string = date_as_string
                result = normaliser.normalise(result)
                if result['domain_info']['name'] == '':
                    result['domain_info']['name'] = domain

        # return the final result
        return result



from .test_unicode_domains import TestUnicodeDomains
from .test_domain_lookups import TestDomainLookups
from .test_dates import TestDates

__all__ = [
    "TestUnicodeDomains",
    "TestDomainLookups",
    "TestDates"
]